"""Delta Lake source for Kukur.

Two formats are supported:
- row based, with many series in one file containing one row per data point
- pivot, with many series as columns in one file
"""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0


try:
    from deltalake import DeltaTable

    HAS_DELTA_LAKE = True
except ImportError:
    HAS_DELTA_LAKE = False

from dataclasses import dataclass
from dataclasses import field as data_field
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import pyarrow as pa
from dateutil.relativedelta import relativedelta

from kukur.base import SeriesSearch, SeriesSelector
from kukur.exceptions import (
    InvalidSourceException,
    MissingModuleException,
)
from kukur.loader import Loader
from kukur.metadata import Metadata
from kukur.source.arrow import (
    cast_ts_column,
    filter_by_timerange,
    filter_pivot_data,
    filter_row_data,
    map_pivot_columns,
    map_row_columns,
)
from kukur.source.quality import QualityMapper


class PartitionOrigin(Enum):
    """Allowed origins for delta lake partitions."""

    TAG = "tag"
    TIMESTAMP = "timestamp"


class Resolution(Enum):
    """Resultion for timestamp partioning."""

    HOUR = "HOUR"
    DAY = "DAY"
    MONTH = "MONTH"
    YEAR = "YEAR"


@dataclass
class DeltaLakePartition:
    """A partition in delta lake."""

    origin: PartitionOrigin
    key: str
    format: Optional[str] = None
    column: Optional[str] = None

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> "DeltaLakePartition":
        """Create a partition from a data dictionary."""
        if "origin" not in data:
            raise InvalidSourceException("No partition origin")
        if "key" not in data:
            raise InvalidSourceException("No partition key")
        return DeltaLakePartition(
            PartitionOrigin(data["origin"]),
            data["key"],
            data.get("format"),
            data.get("column"),
        )


@dataclass
class DeltaSourceOptions:
    """Options for a DeltaSource."""

    data_format: str
    column_mapping: Dict[str, str]
    tag_columns: List[str]
    field_columns: List[str]
    partitions: List[DeltaLakePartition] = data_field(default_factory=list)
    data_datetime_format: Optional[str] = None
    data_timezone: Optional[str] = None
    path_encoding: Optional[str] = None

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> "DeltaSourceOptions":
        """Create source options from a data dictionary."""
        data_format = data.get("format", "row")
        partitions = []
        for partition_data in data.get("partitions", []):
            if "origin" not in partition_data:
                raise InvalidSourceException("No partition origin")
            partitions.append(DeltaLakePartition.from_data(partition_data))

        options = cls(
            data_format,
            data.get("column_mapping", {}),
            data.get("tag_columns", ["series name"]),
            data.get("field_columns", ["value"]),
            partitions,
            data.get("data_datetime_format"),
            data.get("data_timezone"),
            data.get("path_encoding"),
        )

        return options


class DeltaLakeLoader:
    """Fakes a loader for Delta Lake tables.

    It does not really load files, as the other loaders do.
    """

    def __init__(self, config: dict) -> None:
        self.__uri = config["uri"]

    def open(self):
        """Return the URI to connect to."""
        return self.__uri

    def has_child(self, subpath: str) -> bool:
        """Not supported for Delta Lake."""
        raise NotImplementedError()

    def open_child(self, subpath: str):
        """Not supported for Delta Lake."""
        raise NotImplementedError()


class DeltaLakeSource:
    """Connect to a Delta Lake."""

    __loader: Loader
    __options: DeltaSourceOptions
    __quality_mapper: QualityMapper
    __sort_by_timestamp: bool = False

    def __init__(
        self,
        options: DeltaSourceOptions,
        loader: Loader,
        quality_mapper: QualityMapper,
        *,
        sort_by_timestamp: bool = False,
    ):
        """Create a new data source."""
        self.__loader = loader
        self.__options = options
        self.__quality_mapper = quality_mapper
        self.__sort_by_timestamp = sort_by_timestamp

    def search(self, selector: SeriesSearch) -> Generator[SeriesSelector, None, None]:
        """Detect series in the data."""
        if self.__options.data_format == "pivot":
            yield from self._search_pivot(selector.source)

        if self.__options.data_format == "row":
            yield from self._search_row(selector.source)

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Feather currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Read data in one of the predefined formats.

        The complete file will be loaded in an Arrow table during processing.
        """
        data = self.__read_all_data(selector, start_date, end_date)
        if self.__sort_by_timestamp is True:
            data = data.sort_by("ts")
        return filter_by_timerange(data, start_date, end_date)

    def get_file_extension(self) -> str:
        """Delta lakes do not support row-based formats."""
        raise NotImplementedError()

    def __read_all_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        if self.__options.data_format == "pivot":
            all_data = self._read_pivot_data()
            return filter_pivot_data(all_data, selector)

        all_data = self._read_row_partitioned_data(selector, start_date, end_date)
        return filter_row_data(all_data, selector, self.__quality_mapper)

    def _search_pivot(self, source_name: str) -> Generator[SeriesSelector, None, None]:
        all_data = self._read_pivot_data()
        for name in all_data.column_names[1:]:
            yield SeriesSelector(source_name, name)

    def _read_row_data(self) -> pa.Table:
        column_names = (
            self.__options.tag_columns + ["ts"] + self.__options.field_columns
        )
        row_data = self._read_file(self.__loader.open())
        row_data = map_row_columns(
            row_data, column_names, self.__options.column_mapping, self.__quality_mapper
        )
        row_data = cast_ts_column(
            row_data, self.__options.data_datetime_format, self.__options.data_timezone
        )
        return row_data

    def _read_row_partitioned_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        column_names = (
            self.__options.tag_columns + ["ts"] + self.__options.field_columns
        )
        row_data = self._read_partitioned_file(
            self.__loader.open(), selector, start_date, end_date
        )
        row_data = map_row_columns(
            row_data, column_names, self.__options.column_mapping, self.__quality_mapper
        )
        row_data = cast_ts_column(
            row_data, self.__options.data_datetime_format, self.__options.data_timezone
        )
        return row_data

    def _read_pivot_data(self) -> pa.Table:
        all_data = self._read_file(self.__loader.open())
        all_data = map_pivot_columns(self.__options.column_mapping, all_data)
        all_data = cast_ts_column(
            all_data, self.__options.data_datetime_format, self.__options.data_timezone
        )
        return all_data

    def _search_row(self, source_name: str) -> Generator[SeriesSelector, None, None]:
        all_data = self._read_row_data()
        for tags in (
            all_data.group_by(self.__options.tag_columns).aggregate([]).to_pylist()
        ):
            for field_name in self.__options.field_columns:
                yield SeriesSelector(source_name, tags, field_name)

    def _read_file(self, file_like) -> pa.Table:
        return DeltaTable(file_like).to_pyarrow_table()

    def _read_partitioned_file(
        self,
        file_like,
        selector: SeriesSelector,
        start_date: datetime,
        end_date: datetime,
    ) -> pa.Table:
        """Return a PyArrow Table for the Delta Table using the defined partitions."""
        partitions: List[Tuple[str, str, Union[List[str], str]]] = []
        for partition in self.__options.partitions:
            if partition.origin == PartitionOrigin.TAG:
                partitions.append(self._format_tag_partition(partition, selector))
            if partition.origin == PartitionOrigin.TIMESTAMP:
                partitions.append(
                    self._format_timestamp_partition(partition, start_date, end_date)
                )

        return DeltaTable(file_like).to_pyarrow_table(partitions)

    def _format_tag_partition(
        self,
        partition: DeltaLakePartition,
        selector: SeriesSelector,
    ) -> Tuple[str, str, str]:
        column_name = self.__options.column_mapping.get(partition.key, partition.key)
        return (column_name, "=", selector.tags[partition.key])

    def _format_timestamp_partition(
        self,
        partition: DeltaLakePartition,
        start_date: datetime,
        end_date: datetime,
    ) -> Tuple[str, str, List[str]]:
        resolution = Resolution(partition.key)
        column = resolution.value
        if partition.column is not None:
            column = partition.column

        format = partition.format
        partition_values = []
        max_partition_values_date = None

        if resolution == Resolution.YEAR:
            start_date = start_date.replace(month=1, day=1, hour=0, minute=0, second=0)
            interval: Union[relativedelta, timedelta] = relativedelta(years=1)
            if format is None:
                format = "%Y"
        elif resolution == Resolution.MONTH:
            start_date = start_date.replace(day=1, hour=0, minute=0, second=0)
            max_partition_values_date = start_date.replace(year=start_date.year + 1)
            interval = relativedelta(months=1)
        elif resolution == Resolution.DAY:
            start_date = start_date.replace(hour=0, minute=0, second=0)
            max_partition_values_date = start_date + timedelta(days=31)
            interval = relativedelta(days=1)
        elif resolution == Resolution.HOUR:
            start_date = start_date.replace(minute=0, second=0)
            max_partition_values_date = start_date + timedelta(days=1)
            interval = timedelta(hours=1)

        if (
            max_partition_values_date is not None
            and end_date > max_partition_values_date
        ):
            end_date = max_partition_values_date

        while start_date < end_date:
            if format is not None:
                partition_values.append(start_date.strftime(format))
            else:
                if resolution == Resolution.MONTH:
                    partition_values.append(f"{start_date.month}")
                elif resolution == Resolution.DAY:
                    partition_values.append(f"{start_date.day}")
                elif resolution == Resolution.HOUR:
                    partition_values.append(f"{start_date.hour}")
            start_date = start_date + interval

        return (column, "in", partition_values)


def from_config(config: dict, quality_mapper: QualityMapper) -> DeltaLakeSource:
    """Create a new delta lake data source from the given configuration dictionary."""
    if not HAS_DELTA_LAKE:
        raise MissingModuleException("deltalake", "delta")

    data_format = config.get("format", "row")
    partitions = []
    for partition_data in config.get("partitions", []):
        partitions.append(DeltaLakePartition.from_data(partition_data))
    options = DeltaSourceOptions(
        data_format,
        config.get("column_mapping", {}),
        config.get("tag_columns", ["series name"]),
        config.get("field_columns", ["value"]),
        partitions,
    )

    if data_format not in ["row", "pivot"]:
        raise InvalidSourceException(
            'Delta lake sources support only the "row" and "pivot" format.'
        )
    if "uri" not in config:
        raise InvalidSourceException('Delta lake sources require an "uri" entry')
    return DeltaLakeSource(
        options,
        DeltaLakeLoader(config),
        quality_mapper,
        sort_by_timestamp=config.get("sort_by_timestamp", True),
    )



"""Common logic for the file formats supported by pyarrow."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from base64 import b64encode
from dataclasses import dataclass
from datetime import datetime
from pathlib import PurePath
from typing import Any, Dict, Generator, List, Optional

import pyarrow as pa
import pyarrow.compute
import pyarrow.types

from kukur import Metadata, SeriesSelector
from kukur.base import SeriesSearch
from kukur.exceptions import InvalidDataError, InvalidSourceException
from kukur.loader import Loader
from kukur.source.quality import QualityMapper


@dataclass
class SourcePartition:
    """A partition in a data set."""

    origin: str
    key: str
    path_encoding: Optional[str] = None
    format: Optional[str] = None
    column: Optional[str] = None

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> "SourcePartition":
        """Create a partition from a data dictionary."""
        if "origin" not in data:
            raise InvalidSourceException("No partition origin")
        if "key" not in data:
            raise InvalidSourceException("No partition key")
        return SourcePartition(
            data["origin"],
            data["key"],
            data.get("path_encoding"),
            data.get("format"),
            data.get("column"),
        )


@dataclass
class BaseArrowSourceOptions:
    """Options for a BaseArrowSource."""

    data_format: str
    column_mapping: Dict[str, str]
    tag_columns: List[str]
    field_columns: List[str]
    data_datetime_format: Optional[str] = None
    data_timezone: Optional[str] = None
    path_encoding: Optional[str] = None
    partitions: Optional[List[SourcePartition]] = None

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> "BaseArrowSourceOptions":
        """Create source options from a data dictionary."""
        data_format = data.get("format", "row")
        options = cls(
            data_format,
            data.get("column_mapping", {}),
            data.get("tag_columns", ["series name"]),
            data.get("field_columns", ["value"]),
            data.get("data_datetime_format"),
            data.get("data_timezone"),
            data.get("path_encoding"),
        )

        if data_format == "dir":
            options.partitions = [
                SourcePartition.from_data(partition)
                for partition in data.get("partitions", [])
            ]
            if len(options.partitions) == 0:
                options.partitions.append(SourcePartition("tag", "series name", None))
        return options


class BaseArrowSource(ABC):
    """Base class for pyarrow file format data sources."""

    __loader: Loader
    __options: BaseArrowSourceOptions
    __quality_mapper: QualityMapper
    __sort_by_timestamp: bool = False

    def __init__(
        self,
        options: BaseArrowSourceOptions,
        loader: Loader,
        quality_mapper: QualityMapper,
        *,
        sort_by_timestamp: bool = False,
    ):
        """Create a new data source."""
        self.__loader = loader
        self.__options = options
        self.__quality_mapper = quality_mapper
        self.__sort_by_timestamp = sort_by_timestamp

    @abstractmethod
    def read_file(self, file_like) -> pa.Table:
        """Read the given file-like object using the pyarrow format reader."""
        ...

    @abstractmethod
    def get_file_extension(self) -> str:
        """Return the file extension for the supported pyarrow format."""
        ...

    def search(self, selector: SeriesSearch) -> Generator[SeriesSelector, None, None]:
        """Detect series in the data."""
        if self.__options.data_format == "pivot":
            yield from self._search_pivot(selector.source)

        if self.__options.data_format == "row":
            yield from self._search_row(selector.source)

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Feather currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Read data in one of the predefined formats.

        The complete file will be loaded in an Arrow table during processing.
        """
        data = self.__read_all_data(selector)
        if self.__sort_by_timestamp is True:
            data = data.sort_by("ts")

        return filter_by_timerange(data, start_date, end_date)

    def __read_all_data(self, selector: SeriesSelector) -> pa.Table:
        if self.__options.data_format == "pivot":
            all_data = self._read_pivot_data()
            return filter_pivot_data(all_data, selector)

        if self.__options.data_format == "dir":
            return self._read_directory_data(selector)

        all_data = self._read_row_data()
        return filter_row_data(all_data, selector, self.__quality_mapper)

    def _search_pivot(self, source_name: str) -> Generator[SeriesSelector, None, None]:
        all_data = self._read_pivot_data()
        for name in all_data.column_names[1:]:
            yield SeriesSelector(source_name, name)

    def _read_row_data(self) -> pa.Table:
        column_names = (
            self.__options.tag_columns + ["ts"] + self.__options.field_columns
        )
        all_data = self.read_file(self.__loader.open())
        all_data = map_row_columns(
            all_data, column_names, self.__options.column_mapping, self.__quality_mapper
        )
        all_data = cast_ts_column(
            all_data, self.__options.data_datetime_format, self.__options.data_timezone
        )
        return all_data

    def _read_pivot_data(self) -> pa.Table:
        all_data = self.read_file(self.__loader.open())
        all_data = map_pivot_columns(self.__options.column_mapping, all_data)
        all_data = cast_ts_column(
            all_data, self.__options.data_datetime_format, self.__options.data_timezone
        )
        return all_data

    def _search_row(self, source_name: str) -> Generator[SeriesSelector, None, None]:
        all_data = self._read_row_data()
        for tags in (
            all_data.group_by(self.__options.tag_columns).aggregate([]).to_pylist()
        ):
            for field_name in self.__options.field_columns:
                yield SeriesSelector(source_name, tags, field_name)

    def _read_directory_data(self, selector: SeriesSelector) -> pa.Table:
        columns = ["ts", "value"]
        if self.__quality_mapper.is_present():
            columns.append("quality")

        if self.__options.partitions is None:
            raise InvalidSourceException(
                'Sources with `format = "dir"` require at least one partition'
            )

        data_path = None
        for partition in self.__options.partitions:
            partition_path = selector.tags[partition.key]
            if partition.path_encoding is not None:
                if partition.path_encoding == "base64":
                    partition_path = b64encode(partition_path.encode()).decode()
            if data_path is None:
                data_path = PurePath(partition_path)
            else:
                data_path = data_path / partition_path

        data = self.read_file(
            self.__loader.open_child(f"{data_path}.{self.get_file_extension()}")
        )

        data = _map_columns(self.__options.column_mapping, data)
        data = data.rename_columns(columns)
        data = cast_ts_column(
            data, self.__options.data_datetime_format, self.__options.data_timezone
        )
        schema = pa.schema(
            [
                ("ts", pa.timestamp("us", "UTC")),
                ("value", get_value_schema_type(data)),
            ]
        )
        if self.__quality_mapper.is_present():
            schema = schema.append(pa.field("quality", pa.int8()))
            data = data.set_column(
                2, "quality", _map_quality(data["quality"], self.__quality_mapper)
            )
        return data.cast(schema)


def get_value_schema_type(data: pa.Table):
    """Get the type of the value column."""
    value_type = pa.float64()
    if len(data) > 0:
        if pa.types.is_string(data["value"][0].type):
            value_type = pa.string()
    return value_type


def map_pivot_columns(column_mapping: Dict[str, str], data: pa.Table) -> pa.Table:
    """Map pivot format columns according to the mapping definition."""
    ts_column_name = data.column_names[0]
    if "ts" in column_mapping:
        ts_column_name = column_mapping["ts"]

    ts_column = data[ts_column_name]
    return data.drop([ts_column_name]).add_column(0, "ts", ts_column)


def cast_ts_column(
    data: pa.Table, data_datetime_format: Optional[str], data_timezone: Optional[str]
) -> pa.Table:
    """Cast the timestamp column considering format and timezone."""
    if data_datetime_format is not None:
        # pylint: disable=no-member
        data = data.set_column(
            data.column_names.index("ts"),
            "ts",
            pa.compute.strptime(data["ts"], data_datetime_format, "us"),
        )
    if data_timezone is not None:
        # pylint: disable=no-member
        data = data.set_column(
            data.column_names.index("ts"),
            "ts",
            pa.compute.assume_timezone(data["ts"], data_timezone),
        )
    return data


def filter_pivot_data(all_data: pa.Table, selector: SeriesSelector) -> pa.Table:
    """Filter resulting pivot data based on selector tags and field."""
    if selector.name not in all_data.column_names:
        raise InvalidDataError(f'column "{selector.name}" not found')
    data = all_data.select([0, selector.name]).rename_columns(["ts", "value"])
    schema = pa.schema(
        [
            ("ts", pa.timestamp("us", "UTC")),
            ("value", get_value_schema_type(data)),
        ]
    )
    return data.cast(schema)


def filter_row_data(
    all_data: pa.Table, selector: SeriesSelector, quality_mapper: QualityMapper
) -> pa.Table:
    """Filter resulting row data based on selector tags and field."""
    for k, v in selector.tags.items():
        all_data = all_data.filter(pyarrow.compute.equal(all_data[k], pa.scalar(v)))
    filtered_data = pa.Table.from_arrays(
        [
            all_data["ts"],
            all_data[selector.field],
        ],
        ["ts", "value"],
    )
    schema = pa.schema(
        [
            ("ts", pa.timestamp("us", "UTC")),
            ("value", get_value_schema_type(filtered_data)),
        ]
    )
    if quality_mapper.is_present():
        schema = schema.append(pa.field("quality", pa.int8()))
        filtered_data = filtered_data.set_column(
            2, "quality", _map_quality(all_data["quality"], quality_mapper)
        )

    return filtered_data.cast(schema)


def _map_quality(quality_data: pa.Array, quality_mapper: QualityMapper) -> pa.Array:
    return quality_mapper.map_array(quality_data)


def map_row_columns(
    all_data: pa.Table,
    column_names: List[str],
    column_mapping: Dict[str, str],
    quality_mapper: QualityMapper,
) -> pa.Table:
    """Map columns according to the provided column mapping."""
    if quality_mapper.is_present():
        column_names.append("quality")

    data_columns = {}
    if len(column_mapping) > 0:
        data_columns = {
            column_name: column_mapping.get(column_name, column_name)
            for column_name in column_names
        }

    all_data = _map_columns(data_columns, all_data)
    all_data = all_data.rename_columns(column_names)
    return all_data


def _map_columns(column_mapping: Dict[str, str], data: pa.Table) -> pa.Table:
    if len(column_mapping) == 0:
        return data

    return pa.Table.from_pydict(
        {
            column_name: data[data_name]
            for column_name, data_name in column_mapping.items()
        }
    )


def filter_by_timerange(
    all_data: pa.Table, start_date: datetime, end_date: datetime
) -> pa.Table:
    """Remove data before start_date and after end_date."""
    # pylint: disable=no-member
    on_or_after = pyarrow.compute.greater_equal(all_data["ts"], pa.scalar(start_date))
    before = pyarrow.compute.less(all_data["ts"], pa.scalar(end_date))
    return all_data.filter(pyarrow.compute.and_(on_or_after, before))



"""Quality mapper for Kukur data sources."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Any, Dict, List, Union

import pyarrow as pa
import pyarrow.compute as pc


class Quality(Enum):
    """Enumeration of possible Kukur quality flags."""

    BAD = 0
    GOOD = 1


class QualityMapper:
    """QualityMapper maps quality values used in a source to the quality values of Kukur.

    GOOD = 1 and BAD = 0.
    """

    __good_mapping: List[Union[str, int]]

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "QualityMapper":
        """Create a new mapper from a dictionary that maps Kukur quality values to external quality values."""
        mapper = cls()
        for quality, quality_values in config.items():
            if quality == Quality.GOOD.name:
                for value_list in quality_values:
                    if isinstance(value_list, list):
                        if len(value_list) > 1:
                            mapper.add_mapping_range(
                                range(int(value_list[0]), int(value_list[1]) + 1)
                            )
                        mapper.add_mapping(value_list[0])
                        continue
                    mapper.add_mapping(value_list)
        return mapper

    def __init__(self):
        self.__good_mapping = []

    def add_mapping(self, quality_value: Union[str, int]):
        """Add a mapping."""
        self.__good_mapping.append(quality_value)

    def add_mapping_range(self, quality_values: range):
        """Add a mapping range."""
        self.__good_mapping.extend(quality_values)

    def from_source(self, source_quality_value: Union[str, int]) -> int:
        """Map a quality value of a source to the Kukur quality value."""
        if source_quality_value in self.__good_mapping:
            return Quality.GOOD.value
        return Quality.BAD.value

    def map_array(self, array: pa.Array) -> pa.Array:
        """Use the given quality mapping to convert the values in the array to GOOD or BAD."""
        good = pa.scalar(Quality.GOOD.value, pa.int8())
        bad = pa.scalar(Quality.BAD.value, pa.int8())
        # pylint: disable=no-member
        return pc.if_else(pc.is_in(array, pa.array(self.__good_mapping)), good, bad)

    def is_present(self) -> bool:
        """Check if there is a quality mapping present."""
        return len(self.__good_mapping) > 0



"""Base classes for connections to SQL-like sources from Kukur."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import date, datetime, timezone, tzinfo
from decimal import Decimal
from typing import Dict, Generator, List, Optional, Tuple, Union

import dateutil.parser
import pyarrow as pa

from kukur import Dictionary, Metadata, SeriesSearch, SeriesSelector
from kukur.exceptions import KukurException
from kukur.metadata import fields
from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper

logger = logging.getLogger(__name__)


class InvalidMetadataError(KukurException):
    """Raised when the metadata is invalid."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"invalid metadata: {message}")


class InvalidConfigurationError(KukurException):
    """Raised when the source configuration is invalid."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"invalid configuration: {message}")


@dataclass
class SQLConfig:  # pylint: disable=too-many-instance-attributes
    """Configuration settings for a SQL connection."""

    connection_string: str
    tag_columns: List[str]
    query_string_parameters: bool = False
    list_query: Optional[str] = None
    list_columns: List[str] = field(default_factory=list)
    field_columns: Optional[List[str]] = None
    metadata_query: Optional[str] = None
    metadata_columns: List[str] = field(default_factory=list)
    dictionary_query: Optional[str] = None
    data_query: Optional[str] = None
    data_query_datetime_format: Optional[str] = None
    data_timezone: Optional[tzinfo] = None
    data_query_timezone: Optional[tzinfo] = None
    enable_trace_logging: bool = False
    query_timeout_seconds: Optional[int] = None

    @classmethod
    def from_dict(cls, data):
        """Create a new SQL data source from a configuration dict."""
        if "connection_string" in data:
            connection_string = data["connection_string"]
        else:
            with open(data["connection_string_path"], encoding="utf-8") as f:
                connection_string = f.read().strip()

        config = SQLConfig(connection_string, data.get("tag_columns", ["series name"]))

        config.list_query = data.get("list_query")
        if config.list_query is None and "list_query_path" in data:
            with open(data["list_query_path"], encoding="utf-8") as f:
                config.list_query = f.read()
        config.list_columns = data.get("list_columns", [])
        config.field_columns = data.get("field_columns")
        config.metadata_query = data.get("metadata_query")
        if config.metadata_query is None and "metadata_query_path" in data:
            with open(data["metadata_query_path"], encoding="utf-8") as f:
                config.metadata_query = f.read()
        config.metadata_columns = data.get("metadata_columns", [])
        config.dictionary_query = data.get("dictionary_query")
        if config.dictionary_query is None and "dictionary_query_path" in data:
            with open(data["dictionary_query_path"], encoding="utf-8") as f:
                config.dictionary_query = f.read()
        config.data_query = data.get("data_query")
        if config.data_query is None and "data_query_path" in data:
            with open(data["data_query_path"], encoding="utf-8") as f:
                config.data_query = f.read()

        config.data_query_datetime_format = data.get("data_query_datetime_format")
        config.query_string_parameters = data.get("query_string_parameters", False)
        if "data_timezone" in data:
            config.data_timezone = dateutil.tz.gettz(data.get("data_timezone"))
        if "data_query_timezone" in data:
            config.data_query_timezone = dateutil.tz.gettz(
                data.get("data_query_timezone")
            )
        if "enable_trace_logging" in data:
            config.enable_trace_logging = data.get("enable_trace_logging", False)
        if data.get("query_timeout_enable", True):
            config.query_timeout_seconds = data.get("query_timeout_seconds", 0)

        return config


class BaseSQLSource(ABC):
    """A SQL data source.

    Subclasses should implement a connect() method.
    """

    _config: SQLConfig
    _metadata_value_mapper: MetadataValueMapper
    _quality_mapper: QualityMapper

    def __init__(
        self,
        config: SQLConfig,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):
        self._config = config
        self._metadata_value_mapper = metadata_value_mapper
        self._quality_mapper = quality_mapper

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[SeriesSelector, Metadata], None, None]:
        """Search for time series matching the given selector."""
        if self._config.list_query is None:
            return
        if len(self._config.list_columns) == 0:
            for result in self.__search_names(selector):
                yield result
            return
        for metadata in self.__search_metadata(selector):
            yield metadata

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Read metadata from the DB-API connection."""
        metadata = Metadata(selector)
        if self._config.metadata_query is None:
            return metadata
        connection = self.connect()
        cursor = connection.cursor()

        query = self._config.metadata_query
        params = [selector.tags[tag_name] for tag_name in self._config.tag_columns]

        if self._config.query_string_parameters:
            query = query.format(*params)
            params = []
        cursor.execute(query, params)
        row = cursor.fetchone()
        if row:
            for i, name in enumerate(self._config.metadata_columns):
                value = row[i]
                if value is None:
                    continue
                if isinstance(value, str) and value == "":
                    continue
                try:
                    metadata.coerce_field(
                        name, self._metadata_value_mapper.from_source(name, value)
                    )
                except ValueError:
                    pass

        dictionary_name = metadata.get_field(fields.DictionaryName)
        if dictionary_name is not None:
            metadata.set_field(
                fields.Dictionary, self.__query_dictionary(cursor, dictionary_name)
            )
        return metadata

    def get_data(  # noqa: PLR0912
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data using the specified DB-API query."""
        if self._config.data_query is None:
            return pa.Table.from_pydict({"ts": [], "value": [], "quality": []})
        connection = self.connect()
        cursor = connection.cursor()

        query, params = self.__prepare_data_query(selector, start_date, end_date)
        cursor.execute(query, params)

        timestamps = []
        values = []
        qualities = []
        for row in cursor:
            if self._config.enable_trace_logging:
                logger.info(
                    'Data from "%s (%s)" at %s has value %s with type %s',
                    selector.source,
                    selector.name,
                    row[0],
                    row[1],
                    type(row[1]),
                )
            if isinstance(row[0], datetime):
                ts = row[0]
            elif isinstance(row[0], date):
                ts = datetime(
                    row[0].year, row[0].month, row[0].day, tzinfo=timezone.utc
                )
            else:
                ts = dateutil.parser.parse(row[0])
            if self._config.data_timezone:
                ts = ts.replace(tzinfo=self._config.data_timezone)
            ts = ts.astimezone(timezone.utc)
            value = row[1]
            if value is None:
                value = float("nan")
            if isinstance(value, bytes):
                continue
            if isinstance(value, (datetime, date)):
                value = value.isoformat()
            elif isinstance(value, Decimal):
                value = float(value)
            if self._quality_mapper.is_present():
                quality = self._quality_mapper.from_source(row[2])
                qualities.append(quality)
            timestamps.append(ts)
            values.append(value)
        if self._quality_mapper.is_present():
            return pa.Table.from_pydict(
                {"ts": timestamps, "value": values, "quality": qualities}
            )
        return pa.Table.from_pydict({"ts": timestamps, "value": values})

    def __prepare_data_query(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> Tuple[str, List]:
        assert self._config.data_query is not None

        try:
            query = self._config.data_query.format(field=selector.field)
        except (TypeError, IndexError):
            query = self._config.data_query
        params = [selector.tags[tag_name] for tag_name in self._config.tag_columns]
        params.extend(
            [
                self.__format_date(start_date),
                self.__format_date(end_date),
            ]
        )
        if self._config.query_string_parameters:
            query = query.format(*params)
            params = []
        return query, params

    def __search_names(
        self, selector: SeriesSearch
    ) -> Generator[SeriesSelector, None, None]:
        connection = self.connect()
        cursor = connection.cursor()
        cursor.execute(self._config.list_query)

        for tag_values in cursor:
            if len(tag_values) != len(self._config.tag_columns):
                raise InvalidConfigurationError(
                    "number of tag_columns does not match result of list_query"
                )
            tags = {
                tag_name: tag_value
                for tag_name, tag_value in zip(self._config.tag_columns, tag_values)
            }
            if self._config.field_columns is not None:
                for field_name in self._config.field_columns:
                    yield SeriesSelector(selector.source, tags, field_name)
            else:
                yield SeriesSelector(selector.source, tags)

    def __search_metadata(
        self, selector: SeriesSearch
    ) -> Generator[Metadata, None, None]:
        for tag_name in self._config.tag_columns:
            if tag_name not in self._config.list_columns:
                raise InvalidMetadataError(f'tag column "{tag_name}" not found')

        connection = self.connect()
        cursor = connection.cursor()
        dictionary_cursor = None
        if self._config.dictionary_query is not None:
            dictionary_cursor = self.connect().cursor()

        cursor.execute(self._config.list_query)
        tag_column_indices = []

        for row in cursor:
            tags = {}
            for i, name in enumerate(self._config.list_columns):
                if name in self._config.tag_columns:
                    if i not in tag_column_indices:
                        tag_column_indices.append(i)
                    tags[name] = row[i]

            if self._config.field_columns is not None:
                for field_name in self._config.field_columns:
                    series_selector = SeriesSelector(selector.source, tags, field_name)
                    yield self.__get_metadata(
                        row, series_selector, tag_column_indices, dictionary_cursor
                    )
            else:
                series_selector = SeriesSelector(selector.source, tags)
                yield self.__get_metadata(
                    row, series_selector, tag_column_indices, dictionary_cursor
                )

    def __get_metadata(
        self,
        row,
        selector: SeriesSelector,
        tag_column_indices: List[int],
        dictionary_cursor,
    ) -> Metadata:
        metadata = Metadata(selector)
        for i, name in enumerate(self._config.list_columns):
            if i in tag_column_indices:
                continue
            value = row[i]
            if value is None:
                continue
            if isinstance(value, str) and value == "":
                continue
            try:
                metadata.coerce_field(
                    name, self._metadata_value_mapper.from_source(name, value)
                )
            except ValueError:
                pass
        dictionary_name = metadata.get_field(fields.DictionaryName)
        if dictionary_name is not None and dictionary_cursor is not None:
            metadata.set_field(
                fields.Dictionary,
                self.__query_dictionary(dictionary_cursor, dictionary_name),
            )
        return metadata

    def __query_dictionary(self, cursor, dictionary_name: str) -> Optional[Dictionary]:
        if self._config.dictionary_query is None:
            return None
        mapping: Dict[int, str] = {}

        query = self._config.dictionary_query
        params = [dictionary_name]
        if self._config.query_string_parameters:
            query = query.format(*params)
            params = []
        cursor.execute(query, params)
        for row in cursor.fetchall():
            mapping[int(row[0])] = row[1]
        if len(mapping) == 0:
            return None
        return Dictionary(mapping)

    def __format_date(self, date):
        if self._config.data_query_timezone:
            date = date.astimezone(self._config.data_query_timezone).replace(
                tzinfo=None
            )
        if self._config.data_query_datetime_format is not None:
            return date.strftime(self._config.data_query_datetime_format)
        return date

    @abstractmethod
    def connect(self):
        """Create a new PEP-249 connection."""
        ...



"""Kukur source for Arrow IPC streams."""

# SPDX-FileCopyrightText: 2023 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict

import pyarrow as pa
from pyarrow.ipc import open_stream

from kukur.exceptions import InvalidSourceException
from kukur.loader import from_config as loader_from_config
from kukur.source.arrow import BaseArrowSource, BaseArrowSourceOptions
from kukur.source.quality import QualityMapper


class ArrowIPCStreamSource(BaseArrowSource):
    """An Arrow IPC stream data source."""

    def read_file(self, file_like) -> pa.Table:
        """Read an Arrow table from the stream."""
        return open_stream(file_like).read_all()

    def get_file_extension(self) -> str:
        """Return the default .arrows stream file extension."""
        return "arrows"


def from_config(
    config: Dict[str, Any], quality_mapper: QualityMapper
) -> ArrowIPCStreamSource:
    """Create a new IPC Stream data source from the given configuration dictionary."""
    loader = loader_from_config(config, files_as_path=True)
    if "path" not in config:
        raise InvalidSourceException('arrows sources require a "path" entry')
    options = BaseArrowSourceOptions.from_data(config)
    return ArrowIPCStreamSource(
        options,
        loader,
        quality_mapper,
        sort_by_timestamp=config.get("sort_by_timestamp", False),
    )



"""Connections to ODBC data sources from Kukur."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

try:
    import pyodbc

    HAS_ODBC = True
except ImportError:
    HAS_ODBC = False

from kukur.exceptions import MissingModuleException
from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper
from kukur.source.sql import BaseSQLSource, SQLConfig


def from_config(
    data, metadata_value_mapper: MetadataValueMapper, quality_mapper: QualityMapper
):
    """Create a new ODBC data source from a configuration dict."""
    if not HAS_ODBC:
        raise MissingModuleException("pyodbc", "odbc")

    config = SQLConfig.from_dict(data)

    return ODBCSource(config, metadata_value_mapper, quality_mapper)


class ODBCSource(BaseSQLSource):
    """An ODBC data source."""

    def __init__(
        self,
        config: SQLConfig,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):
        super().__init__(config, metadata_value_mapper, quality_mapper)
        if not HAS_ODBC:
            raise MissingModuleException("pyodbc", "odbc")

    def connect(self):
        """Return a pyodbc connection."""
        connection = pyodbc.connect(self._config.connection_string)
        if self._config.query_timeout_seconds is not None:
            connection.timeout = self._config.query_timeout_seconds
        return connection



"""PostgreSQL connection for Kukur.

This uses the LGPL psycopg library.
"""

# SPDX-FileCopyrightText: 2023 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0


try:
    import psycopg

    HAS_POSTGRES = True
except ImportError:
    HAS_POSTGRES = False

from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper
from kukur.source.sql import BaseSQLSource, SQLConfig


class PostgresSource(BaseSQLSource):
    """Kukur source for PostgreSQL."""

    def connect(self):
        """Create a connection to PostgreSQL."""
        return psycopg.connect(self._config.connection_string)


def from_config(
    data, metadata_value_mapper: MetadataValueMapper, quality_mapper: QualityMapper
) -> PostgresSource:
    """Create a new PostgresSource source from a configuration dictionay."""
    config = SQLConfig.from_dict(data)
    return PostgresSource(config, metadata_value_mapper, quality_mapper)



"""csv contains the CSV data source for Timeseer.

Three formats are supported:
- row based, with many series in one file containing one line per data point
- directory based, with one file per series
- pivot, with many series as columns in one file
"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import csv
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Union

import pyarrow as pa
import pyarrow.compute
import pyarrow.csv

from kukur import Dictionary, Metadata, SeriesSearch, SeriesSelector
from kukur.exceptions import InvalidDataError, InvalidSourceException, KukurException
from kukur.loader import Loader
from kukur.loader import from_config as loader_from_config
from kukur.metadata import fields
from kukur.source.metadata import MetadataMapper, MetadataValueMapper
from kukur.source.quality import QualityMapper


class InvalidMetadataError(KukurException):
    """Raised when the metadata is invalid."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"invalid metadata: {message}")


def from_config(
    config: Dict[str, Any],
    metadata_mapper: MetadataMapper,
    metadata_value_mapper: MetadataValueMapper,
    quality_mapper: QualityMapper,
):
    """Create a new CSV data source from the given configuration dictionary."""
    loaders = CSVLoaders()
    if "path" in config:
        loaders.data = loader_from_config(config, "path", files_as_path=True)
    if "metadata" in config:
        loaders.metadata = loader_from_config(config, "metadata", "r")
    if "dictionary_dir" in config:
        loaders.dictionary = loader_from_config(config, "dictionary_dir", "r")
    data_format = config.get("format", "row")
    column_mapping = config.get("column_mapping", {})
    data_datetime_format = config.get("data_datetime_format", None)
    data_timezone = config.get("data_timezone", None)
    options = CSVSourceOptions(
        data_format,
        config.get("header_row", False),
        column_mapping,
        config.get("tag_columns", ["series name"]),
        config.get("field_columns", ["value"]),
        config.get("metadata_field_column"),
        data_datetime_format,
        data_timezone,
    )
    metadata_fields: List[str] = config.get("metadata_fields", [])
    if len(metadata_fields) == 0:
        metadata_fields = config.get("fields", [])
    mappers = CSVMappers(metadata_mapper, metadata_value_mapper, quality_mapper)
    return CSVSource(options, metadata_fields, loaders, mappers)


@dataclass
class CSVLoaders:
    """Data loaders for CSV sources."""

    data: Optional[Loader] = None
    metadata: Optional[Loader] = None
    dictionary: Optional[Loader] = None


@dataclass
class CSVMappers:
    """Value mappers for CSV sources."""

    metadata: MetadataMapper
    metadata_values: MetadataValueMapper
    quality: QualityMapper


@dataclass
class CSVSourceOptions:
    """Options for a CSV source."""

    data_format: str
    header_row: bool
    column_mapping: Dict[str, str]
    tags: List[str]
    fields: List[str]
    metadata_field_column: Optional[str] = None
    data_datetime_format: Optional[str] = None
    data_timezone: Optional[str] = None


class CSVSource:
    """A CSV data source."""

    __loaders: CSVLoaders
    __options: CSVSourceOptions
    __metadata_fields: List[str]
    __mappers: CSVMappers

    def __init__(
        self,
        options: CSVSourceOptions,
        metadata_fields: List[str],
        loaders: CSVLoaders,
        mappers: CSVMappers,
    ):
        """Create a new CSV data source."""
        self.__loaders = loaders
        self.__metadata_fields = metadata_fields
        self.__options = options
        self.__mappers = mappers

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Search for series matching the given selector."""
        if self.__loaders.metadata is None:
            yield from self._search_in_data(selector)
            return
        with self.__loaders.metadata.open() as metadata_file:
            reader = csv.DictReader(metadata_file)
            for row in reader:
                tags = {}
                for tag in self.__options.tags:
                    column_name = self._map_column_name(tag)
                    if column_name not in row:
                        raise InvalidMetadataError('column "{column_name}" not found')
                    tags[tag] = row[column_name]

                field = None
                if self.__options.metadata_field_column is not None:
                    column_name = self._map_column_name(
                        self.__options.metadata_field_column
                    )
                    if column_name not in row:
                        raise InvalidMetadataError('column "{column_name}" not found')
                    field = row[column_name]

                metadata = Metadata(
                    SeriesSelector.from_tags(selector.source, tags, field)
                )

                field_names = [field for field, _ in metadata.iter_names()]
                if len(self.__metadata_fields) > 0:
                    field_names = self.__metadata_fields
                for field in field_names:
                    column_name = self._map_column_name(field)
                    if column_name in row:
                        value = row[column_name]
                        try:
                            metadata.coerce_field(
                                field,
                                self.__mappers.metadata_values.from_source(
                                    field, value
                                ),
                            )
                        except ValueError:
                            pass
                dictionary_name = metadata.get_field(fields.DictionaryName)
                if dictionary_name is not None:
                    metadata.set_field(
                        fields.Dictionary, self.__get_dictionary(dictionary_name)
                    )
                yield metadata

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Read metadata, taking any configured metadata mapping into account."""
        metadata = Metadata(selector)
        if self.__loaders.metadata is None:
            return metadata

        with self.__loaders.metadata.open() as metadata_file:
            reader = csv.DictReader(metadata_file)
            for row in reader:
                skip_row = False
                for tag in self.__options.tags:
                    column_name = self._map_column_name(tag)
                    if column_name not in row:
                        raise InvalidMetadataError(f'column "{column_name}" not found')
                    if row[column_name] != selector.tags[tag]:
                        skip_row = True
                if skip_row:
                    continue
                if self.__options.metadata_field_column is not None:
                    column_name = self._map_column_name(
                        self.__options.metadata_field_column
                    )
                    if column_name in row:
                        if row[column_name] != selector.field:
                            continue

                field_names = [field for field, _ in metadata.iter_names()]
                if len(self.__metadata_fields) > 0:
                    field_names = self.__metadata_fields
                for field in field_names:
                    mapped_field = self._map_column_name(field)
                    if mapped_field in row:
                        try:
                            value = row[mapped_field]
                            metadata.coerce_field(
                                field,
                                self.__mappers.metadata_values.from_source(
                                    field, value
                                ),
                            )
                        except ValueError:
                            pass

            dictionary_name = metadata.get_field(fields.DictionaryName)
            if dictionary_name is not None:
                metadata.set_field(
                    fields.Dictionary, self.__get_dictionary(dictionary_name)
                )

        return metadata

    def _map_column_name(self, column_name: str) -> str:
        column_name = self.__mappers.metadata.from_kukur(column_name)
        return self.__options.column_mapping.get(column_name, column_name)

    def __get_dictionary(self, set_name: str) -> Optional[Dictionary]:
        if self.__loaders.dictionary is None:
            return None
        if not self.__loaders.dictionary.has_child(f"{set_name}.csv"):
            return None
        with self.__loaders.dictionary.open_child(f"{set_name}.csv") as mapping_file:
            reader = csv.reader(mapping_file)
            mapping = {int(row[0]): row[1] for row in reader}
            return Dictionary(mapping)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Read data in one of the predefined formats.

        The complete CSV file will be loaded in an Arrow table during processing.
        """
        data = self.__read_all_data(selector)
        # pylint: disable=no-member
        on_or_after = pyarrow.compute.greater_equal(data["ts"], pa.scalar(start_date))
        before = pyarrow.compute.less(data["ts"], pa.scalar(end_date))
        return data.filter(pyarrow.compute.and_(on_or_after, before))

    def _search_in_data(
        self, search: SeriesSearch
    ) -> Generator[SeriesSelector, None, None]:
        if self.__loaders.data is None:
            return

        if self.__options.data_format == "row":
            yield from self._search_row(self.__loaders.data, search)
        elif self.__options.data_format == "pivot":
            yield from self._search_pivot(self.__loaders.data, search)

    def _search_row(
        self, loader: Loader, search: SeriesSearch
    ) -> Generator[SeriesSelector, None, None]:
        all_data = self._open_row_data(loader)
        for tags in all_data.group_by(self.__options.tags).aggregate([]).to_pylist():
            for field_name in self.__options.fields:
                yield SeriesSelector(search.source, tags, field_name)

    def _search_pivot(
        self, loader: Loader, search: SeriesSearch
    ) -> Generator[SeriesSelector, None, None]:
        all_data = self._open_pivot_data(loader)
        for name in all_data.column_names[1:]:
            yield SeriesSelector(search.source, name)

    def __read_all_data(self, selector: SeriesSelector) -> pa.Table:
        if self.__loaders.data is None:
            raise InvalidSourceException("No data path configured.")
        if self.__options.data_format == "pivot":
            return self._read_pivot_data(self.__loaders.data, selector)

        if self.__options.data_format == "dir":
            return self._read_directory_data(self.__loaders.data, selector)

        return self._read_row_data(self.__loaders.data, selector)

    def _read_row_data(self, loader: Loader, selector: SeriesSelector) -> pa.Table:
        row_data = self._open_row_data(loader)
        # pylint: disable=no-member
        for k, v in selector.tags.items():
            row_data = row_data.filter(pyarrow.compute.equal(row_data[k], pa.scalar(v)))
        data = pa.Table.from_arrays(
            [
                row_data["ts"],
                row_data[selector.field],
            ],
            ["ts", "value"],
        )
        if self.__mappers.quality.is_present():
            data = data.set_column(2, "quality", row_data["quality"])
        return data

    def _open_row_data(self, loader: Loader) -> pa.Table:
        columns = self.__options.tags + ["ts"] + self.__options.fields
        timestamp_column = self.__options.column_mapping.get("ts", "ts")

        if self.__mappers.quality.is_present():
            columns.append("quality")

        data_columns = {
            column_name: self.__options.column_mapping.get(column_name, column_name)
            for column_name in columns
        }

        if not self.__options.header_row:
            read_options = pyarrow.csv.ReadOptions(column_names=columns)
        else:
            read_options = pyarrow.csv.ReadOptions()

        convert_options = _get_convert_options(
            timestamp_column,
            self.__options.data_datetime_format,
            self.__options.data_timezone,
        )

        try:
            all_data = pyarrow.csv.read_csv(
                loader.open(),
                read_options=read_options,
                convert_options=convert_options,
            )

            all_data = _map_columns(data_columns, all_data)
            all_data = _cast_ts_column(all_data, self.__options.data_timezone)
        except pa.lib.ArrowInvalid as arrow_invalid_exception:
            if self.__options.data_datetime_format is None:
                raise arrow_invalid_exception

            column_types = {timestamp_column: pa.string()}
            convert_options = pyarrow.csv.ConvertOptions(
                column_types=column_types,
            )
            all_data = pyarrow.csv.read_csv(
                loader.open(),
                read_options=read_options,
                convert_options=convert_options,
            )
            all_data = _map_columns(data_columns, all_data)
            all_data = _convert_timestamp(
                all_data,
                self.__options.data_datetime_format,
                self.__options.data_timezone,
            )

        if self.__mappers.quality.is_present():
            for i, name in enumerate(all_data.column_names):
                if name == "quality":
                    all_data = all_data.set_column(
                        i, "quality", self._map_quality(all_data["quality"])
                    )
        return all_data

    def _read_directory_data(
        self, loader: Loader, selector: SeriesSelector
    ) -> pa.Table:
        columns = ["ts", "value"]
        timestamp_column = "ts"
        if "ts" in self.__options.column_mapping:
            timestamp_column = self.__options.column_mapping["ts"]

        if self.__mappers.quality.is_present():
            columns.append("quality")

        data_columns = {
            column_name: self.__options.column_mapping.get(column_name, column_name)
            for column_name in columns
        }

        if not self.__options.header_row:
            read_options = pyarrow.csv.ReadOptions(column_names=columns)
        else:
            read_options = pyarrow.csv.ReadOptions()

        convert_options = _get_convert_options(
            "ts", self.__options.data_datetime_format, self.__options.data_timezone
        )

        try:
            data = pyarrow.csv.read_csv(
                loader.open_child(f"{selector.tags['series name']}.csv"),
                read_options=read_options,
                convert_options=convert_options,
            )
            data = _map_columns(data_columns, data)
            data = _cast_ts_column(data, self.__options.data_timezone)
        except pa.lib.ArrowInvalid as arrow_invalid_exception:
            if self.__options.data_datetime_format is None:
                raise arrow_invalid_exception

            column_types = {timestamp_column: pa.string()}
            convert_options = pyarrow.csv.ConvertOptions(
                column_types=column_types,
            )
            data = pyarrow.csv.read_csv(
                loader.open_child(f"{selector.tags['series name']}.csv"),
                read_options=read_options,
                convert_options=convert_options,
            )
            data = _map_columns(data_columns, data)
            data = _convert_timestamp(
                data,
                self.__options.data_datetime_format,
                self.__options.data_timezone,
            )
        if self.__mappers.quality.is_present():
            return data.set_column(2, "quality", self._map_quality(data["quality"]))
        return data

    def _read_pivot_data(self, loader: Loader, selector: SeriesSelector) -> pa.Table:
        all_data = self._open_pivot_data(loader)
        if selector.name not in all_data.column_names:
            raise InvalidDataError(f'column "{selector.name}" not found')
        return all_data.select(["ts", selector.name]).rename_columns(["ts", "value"])

    def _open_pivot_data(self, loader: Loader) -> pa.Table:
        timestamp_column = "ts"
        if "ts" in self.__options.column_mapping:
            timestamp_column = self.__options.column_mapping["ts"]

        convert_options = _get_convert_options(
            timestamp_column,
            self.__options.data_datetime_format,
            self.__options.data_timezone,
        )
        try:
            all_data = pyarrow.csv.read_csv(
                loader.open(), convert_options=convert_options
            )
            all_data = _map_pivot_columns(self.__options.column_mapping, all_data)
            all_data = _cast_ts_column(all_data, self.__options.data_timezone)
        except pa.lib.ArrowInvalid as arrow_invalid_exception:
            if self.__options.data_datetime_format is None:
                raise arrow_invalid_exception

            column_types = {timestamp_column: pa.string()}
            convert_options = pyarrow.csv.ConvertOptions(
                column_types=column_types,
            )
            all_data = pyarrow.csv.read_csv(
                loader.open(),
                convert_options=convert_options,
            )
            all_data = _map_pivot_columns(self.__options.column_mapping, all_data)
            all_data = _convert_timestamp(
                all_data,
                self.__options.data_datetime_format,
                self.__options.data_timezone,
            )
        return all_data

    def _map_quality(self, quality_data: pa.Array) -> pa.Array:
        return self.__mappers.quality.map_array(quality_data)


def _get_convert_options(
    timestamp_column: str,
    data_datetime_format: Optional[str],
    data_timezone: Optional[str],
) -> pyarrow.csv.ConvertOptions:
    column_types = {
        timestamp_column: pa.timestamp("us")
        if data_timezone is not None
        else pa.timestamp("us", "UTC")
    }

    timestamp_parsers = (
        [data_datetime_format] if data_datetime_format is not None else None
    )
    return pyarrow.csv.ConvertOptions(
        column_types=column_types,
        timestamp_parsers=timestamp_parsers,
    )


def _cast_ts_column(data: pa.Table, data_timezone: Optional[str]) -> pa.Table:
    if data_timezone is None:
        return data

    # pylint: disable=no-member
    return data.set_column(
        data.column_names.index("ts"),
        "ts",
        pyarrow.compute.assume_timezone(data["ts"], data_timezone),
    )


def _convert_timestamp(
    data: pa.Table, data_datetime_format: str, data_timezone: Optional[str]
) -> pa.Table:
    # pylint: disable=no-member
    data = data.set_column(
        data.column_names.index("ts"),
        "ts",
        [
            [
                datetime.strptime(timestamp.as_py(), data_datetime_format)  # noqa: DT
                for timestamp in data["ts"]
            ]
        ],
    )

    if data_timezone is None:
        if data["ts"][0].as_py().tzinfo is None:
            # pylint: disable=no-member
            return data.set_column(
                data.column_names.index("ts"),
                "ts",
                pyarrow.compute.assume_timezone(data["ts"], "UTC"),
            )
        return data

    # pylint: disable=no-member
    return data.set_column(
        data.column_names.index("ts"),
        "ts",
        pyarrow.compute.assume_timezone(data["ts"], data_timezone),
    )


def _map_columns(column_mapping: Dict[str, str], data: pa.Table) -> pa.Table:
    return pa.Table.from_pydict(
        {
            column_name: data[data_name]
            for column_name, data_name in column_mapping.items()
        }
    )


def _map_pivot_columns(column_mapping: Dict[str, str], data: pa.Table) -> pa.Table:
    ts_column_name = data.column_names[0]
    if "ts" in column_mapping:
        ts_column_name = column_mapping["ts"]

    ts_column = data[ts_column_name]
    return data.drop([ts_column_name]).add_column(0, "ts", ts_column)



"""Kukur source for JSON files."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import json
from datetime import datetime
from pathlib import Path
from typing import Dict, Generator

from pyarrow import Table

from kukur import Metadata, SeriesSearch, SeriesSelector
from kukur.exceptions import InvalidSourceException


class JSONSource:
    """Read Kukur exported JSON data."""

    __path: Path

    def __init__(self, path: Path):
        self.__path = path

    def search(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Read all Kukur JSON files with metadata."""
        for metadata_file in self.__path.glob("*.json"):
            series_name = metadata_file.stem
            yield _read_metadata(
                SeriesSelector(selector.source, series_name), metadata_file
            )

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Read exactly one Kukur JSON file."""
        path = self.__path / (
            self.__path.joinpath(f"{selector.tags['series name']}.json")
            .resolve()
            .relative_to(self.__path.resolve())
        )
        metadata = Metadata(selector)
        if not path.is_file():
            return metadata
        return _read_metadata(selector, path)

    def get_data(
        self, _selector: SeriesSelector, _start_date: datetime, _end_date: datetime
    ) -> Table:
        """Get data from the Flight service."""
        raise InvalidSourceException("The 'json' source does not support reading data.")


def _read_metadata(selector: SeriesSelector, path: Path) -> Metadata:
    with path.open("rb") as f:
        data = json.load(f)
        return Metadata.from_data(data, selector)


def from_config(config: Dict) -> JSONSource:
    """Create a new JSON source."""
    if "metadata" not in config:
        raise InvalidSourceException("'metadata' is required for 'json' sources")
    path = Path(config["metadata"])
    if not path.is_dir():
        raise InvalidSourceException("'metadata' should be a directory")
    return JSONSource(path)



"""parquet contains the Parquet data source for Timeseer.

Three formats are supported:
- row based, with many series in one file containing one row per data point
- directory based, with one file per series
- pivot, with many series as columns in one file

Parquet DataSets are not yet supported.
"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict

import pyarrow as pa
from pyarrow import parquet

from kukur.exceptions import InvalidSourceException
from kukur.loader import from_config as loader_from_config
from kukur.source.arrow import BaseArrowSource, BaseArrowSourceOptions
from kukur.source.quality import QualityMapper


def from_config(config: Dict[str, Any], quality_mapper: QualityMapper):
    """Create a new Parquet data source from the given configuration dictionary."""
    loader = loader_from_config(config, files_as_path=True)
    if "path" not in config:
        raise InvalidSourceException('Parquet sources require a "path" entry')
    options = BaseArrowSourceOptions.from_data(config)
    return ParquetSource(options, loader, quality_mapper)


class ParquetSource(BaseArrowSource):
    """A Parquet data source."""

    def read_file(self, file_like) -> pa.Table:
        """Read the file_like object as Parquet."""
        return parquet.read_table(file_like)

    def get_file_extension(self) -> str:
        """Return the default parquet file extension."""
        return "parquet"



"""Package Kukur for publishing."""

import os

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kukur",
    version=os.environ.get("KUKUR_VERSION", "0.0.0"),
    author="Timeseer.AI",
    author_email="pypi@timeseer.ai",
    description="Kukur makes time series data and metadata available to the Apache Arrow ecosystem.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    url="https://kukur.timeseer.ai/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=setuptools.find_packages(exclude="tests"),
    install_requires=[
          "pyarrow",
          "python-dateutil",
          "pytz",
          "toml",
      ],
    extras_require={
        "adodb": ["pywin32"],
        "cratedb": ["crate"],
        "delta": ["deltalake"],
        "influxdb": ["influxdb"],
        "kusto": ["azure-identity", "azure-kusto-data"],
        "odbc": ["pyodbc"],
        "piwebapi": ["requests", "requests-kerberos"],
        "postgresql": ["psycopg[binary]"],
    },
    python_requires=">=3.8",
    package_data={
        "kukur": ["py.typed"],
    },
    entry_points={
        "console_scripts": ["kukur = kukur.cli:_run"],
    },
)



"""Tests for SeriesSelector."""

# SPDX-FileCopyrightText: 2023 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from kukur import SeriesSelector


def test_name_series_name() -> None:
    selector = SeriesSelector("source", {"series name": "a"}, "field")
    assert selector.name == "a::field"
    assert selector == SeriesSelector.from_name("source", selector.name)


def test_name_series_name_value() -> None:
    selector = SeriesSelector("source", {"series name": "a"}, "value")
    assert selector.name == "a"
    assert selector == SeriesSelector.from_name("source", selector.name)


def test_name() -> None:
    selector = SeriesSelector("source", {"tag-a": "a"}, "field")
    assert selector.name == "tag-a=a::field"
    assert selector == SeriesSelector.from_name("source", selector.name)


def test_name_value() -> None:
    selector = SeriesSelector("source", {"tag-a": "a"}, "value")
    assert selector.name == "tag-a=a"
    assert selector == SeriesSelector.from_name("source", selector.name)


def test_name_multiple_tags() -> None:
    selector = SeriesSelector("source", {"tag-a": "a", "tag-b": "b"}, "field")
    assert selector.name == "tag-a=a,tag-b=b::field"
    assert selector == SeriesSelector.from_name("source", selector.name)


def test_name_multiple_tags_series_name_first() -> None:
    selector = SeriesSelector(
        "source", {"tag-a": "a", "tag-b": "b", "series name": "c"}, "field"
    )
    assert selector.name == "c,tag-a=a,tag-b=b::field"
    assert selector == SeriesSelector.from_name("source", selector.name)


def test_from_name_strip_whitespace() -> None:
    selector = SeriesSelector(
        "source", {"tag-a": "a", "tag-b": "b", "series name": "c"}, "field"
    )
    assert selector == SeriesSelector.from_name("source", " c,tag-a=a,tag-b=b::field ")
